import logging
import json
from functools import wraps, update_wrapper, partialmethod

import tornado.web
import tornado.httpclient
from tornado.platform.asyncio import to_asyncio_future

from rest_tools.client import RestClient
from rest_tools.utils import Auth
from rest_tools.server import RestHandlerSetup, authenticated, catch_error
from rest_tools.server import RestHandler as BaseRestHandler

import iceprod

logger = logging.getLogger('rest')

def RESTHandlerSetup(config, module=None):
    ret = RestHandlerSetup(config)
    ret['module'] = module
    return ret

class RESTHandler(BaseRestHandler):
    """Default REST handler"""
    def initialize(self, module=None, **kwargs):
        super(RESTHandler, self).initialize(**kwargs)
        self.module = module

    def prepare(self):
        if self.module and self.module.statsd:
            self.module.statsd.incr('prepare.{}.{}'.format(self.__class__.__name__, self.request.method))
            # for individual dataset/job/task info
            #self.module.statsd.incr('prepare.{}.{}'.format(self.request.path.replace('.','_'), self.request.method))

    def on_finish(self):
        if self.module and self.module.statsd:
            self.module.statsd.incr('finish.{}.{}.{}'.format(self.__class__.__name__,
                             self.request.method, self.get_status()))

    def set_default_headers(self):
        self._headers['Server'] = 'IceProd/' + iceprod.__version__

    def get_template_namespace(self):
        namespace = super(RESTHandler,self).get_template_namespace()
        namespace['version'] = iceprod.__version__
        return namespace



def authorization(**_auth):
    """
    Handle authorization.

    Like :py:func:`iceprod.server.rest.authenticated`, this requires the Authorization header
    to be filled with a valid token.  Note that calling both decorators
    is not necessary, as this decorator will perform authentication
    checking as well.

    Args:
        roles (list): The roles to match
        attrs (list): The attributes to match

    Raises:
        :py:class:`tornado.web.HTTPError`
    """
    def make_wrapper(method):
        @authenticated
        @catch_error
        @wraps(method)
        async def wrapper(self, *args, **kwargs):
            roles = _auth.get('roles', [])
            attrs = _auth.get('attrs', {})

            authorized = False

            auth_role = self.auth_data.get('role',None)
            if roles and auth_role in roles:
                authorized = True
            else:
                logger.info('roles: %r    attrs: %r', roles, attrs)
                logger.info('token_role: %r', auth_role)
                logger.info('role mismatch')

            if (not authorized) and ('dataset_id:read' in attrs or 'dataset_id:write' in attrs):
                # we need to ask /auths about this
                try:
                    dataset_id = kwargs.get('dataset_id', None)
                    if (not dataset_id) or not isinstance(dataset_id,str):
                        raise tornado.web.HTTPError(403, reason="authorization failed")
                    url = '/auths/'+dataset_id+'/actions/'
                    http_client = RestClient(self.auth_url, token=self.auth_key)
                    if 'dataset_id:read' in attrs:
                        await http_client.request('GET', url+'read')
                        authorized = True
                    elif 'dataset_id:write' in attrs:
                        await http_client.request('GET', url+'write')
                        authorized = True
                except Exception:
                    logger.info('/auths failure', exc_info=True)
                    raise tornado.web.HTTPError(403, reason="authorization failed")

            if authorized:
                return await method(self, *args, **kwargs)
            else:
                raise tornado.web.HTTPError(403, reason="authorization failed")
        return wrapper
    return make_wrapper



"""
The website module uses `Tornado <http://www.tornadoweb.org>`_,
a fast and scalable python web server.

Main website
------------

This is the external website users will see when interacting with IceProd.
It has been broken down into several sub-handlers for easier maintenance.

"""
from __future__ import absolute_import, division, print_function

import sys
import os
import time
import random
import binascii
import socket
from threading import Thread,Event,Condition
import logging
from contextlib import contextmanager
from functools import partial, wraps
try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse
from datetime import timedelta
from collections import defaultdict

from iceprod.core.jsonUtil import json_encode,json_decode

import tornado.ioloop
import tornado.web
import tornado.httpserver
import tornado.gen

import tornado.concurrent
import concurrent.futures
from rest_tools.client import RestClient
from rest_tools.server import RestServer

import iceprod
from iceprod.server import GlobalID, get_pkgdata_filename
from iceprod.server import module
from iceprod.server.ssl_cert import create_cert, verify_cert
import iceprod.core.functions
from iceprod.server import documentation

logger = logging.getLogger('website')

class website(module.module):
    """
    The main website module.

    Run the website, which is required for anything to work.
    """

    def __init__(self,*args,**kwargs):
        # run default init
        super(website,self).__init__(*args,**kwargs)

        # set up local variables
        self.http_server = None

    async def stop(self):
        """Stop website"""
        # stop tornado
        try:
            if self.http_server:
                await self.http_server.stop()
        except Exception:
            logger.error('cannot stop tornado', exc_info=True)
        super(website,self).stop()

    def start(self):
        """Run the website"""
        super(website,self).start()

        try:
            # make sure directories are set up properly
            for d in self.cfg['webserver']:
                if '_dir' in d:
                    path = self.cfg['webserver'][d]
                    path = os.path.expanduser(os.path.expandvars(path))
                    try:
                        os.makedirs(path)
                    except Exception:
                        pass

            # get package data
            static_path = get_pkgdata_filename('iceprod.server','data/www')
            if static_path is None or not os.path.exists(static_path):
                logger.info('static path: %r',static_path)
                raise Exception('bad static path')
            template_path = get_pkgdata_filename('iceprod.server','data/www_templates')
            if template_path is None or not os.path.exists(template_path):
                logger.info('template path: %r',template_path)
                raise Exception('bad template path')

            if 'url' in self.cfg['rest_api']:
                rest_address = self.cfg['rest_api']['url']
            else:
                # for local systems
                rest_address = 'http://{}:{}'.format(
                        self.cfg['rest_api']['address'],
                        self.cfg['rest_api']['port'],
                )

            handler_args = {
                'cfg':self.cfg,
                'modules':self.modules,
                'statsd':self.statsd,
                'rest_api':rest_address,
                'debug':True,
            }
            login_handler_args = handler_args.copy()
            login_handler_args['module_rest_client'] = self.rest_client
            if 'debug' in self.cfg['webserver'] and self.cfg['webserver']['debug']:
                handler_args['debug'] = True
            if 'cookie_secret' in self.cfg['webserver']:
                cookie_secret = self.cfg['webserver']['cookie_secret']
            else:
                cookie_secret = ''.join(hex(random.randint(0,15))[-1] for _ in range(64))
                self.cfg['webserver']['cookie_secret'] = cookie_secret

            routes = [
                (r"/", Default, handler_args),
                (r"/submit", Submit, handler_args),
                (r"/config", Config, handler_args),
                (r"/dataset", DatasetBrowse, handler_args),
                (r"/dataset/(\w+)", Dataset, handler_args),
                (r"/dataset/(\w+)/task", TaskBrowse, handler_args),
                (r"/dataset/(\w+)/task/(\w+)", Task, handler_args),
                (r"/dataset/(\w+)/job", JobBrowse, handler_args),
                (r"/dataset/(\w+)/job/(\w+)", Job, handler_args),
                (r"/help", Help, handler_args),
                (r"/docs/(.*)", Documentation, handler_args),
                (r"/dataset/(\w+)/log/(\w+)", Log, handler_args),
                #(r"/groups", GroupsHandler, handler_args),
                (r'/profile', Profile, handler_args),
                (r"/login", Login, login_handler_args),
                (r"/logout", Logout, handler_args),
                (r"/.*", Other, handler_args),
            ]
            self.http_server = RestServer(
                static_path=static_path,
                template_path=template_path,
                cookie_secret=cookie_secret,
                login_url='/login',
                debug=handler_args['debug'],
            )
            for r in routes:
                self.http_server.add_route(*r)

            # start tornado
            self.http_server.startup(
                port=self.cfg['webserver']['port'],
                address='0.0.0.0', # bind to all
            )
        except Exception:
            logger.error('website startup error',exc_info=True)
            raise


def catch_error(method):
    """Decorator to catch and handle errors on handlers"""
    @wraps(method)
    def wrapper(self, *args, **kwargs):
        try:
            return method(self, *args, **kwargs)
        except Exception as e:
            self.statsd.incr(self.__class__.__name__+'.error')
            logger.warning('Error in website handler', exc_info=True)
            message = 'Error generating page for '+self.__class__.__name__
            if self.debug:
                message = message + '\n' + str(e)
            self.send_error(500, message=message)
    return wrapper

def authenticated_secure(method):
    """Decorate methods with this to require that the user be logged in
    to a secure area.

    If the user is not logged in, they will be redirected to the configured
    `login url <RequestHandler.get_login_url>`.

    If you configure a login url with a query parameter, Tornado will
    assume you know what you're doing and use it as-is.  If not, it
    will add a `next` parameter so the login page knows where to send
    you once you're logged in.
    """
    @wraps(method)
    def wrapper(self, *args, **kwargs):
        if not self.current_user_secure:
            if self.request.method in ("GET", "HEAD"):
                url = self.get_login_url()
                if "?" not in url:
                    if urlparse.urlsplit(url).scheme:
                        # if login url is absolute, make next absolute too
                        next_url = self.request.full_url()
                    else:
                        next_url = self.request.uri
                    url += "?" + urlencode({'next':next_url,'secure':True})
                self.redirect(url)
                return
            raise HTTPError(403)
        return method(self, *args, **kwargs)
    return wrapper


class PublicHandler(tornado.web.RequestHandler):
    """Default Handler"""
    def initialize(self, cfg, modules, debug=False, statsd=None,
                   rest_api=None):
        """
        Get some params from the website module

        :param cfg: the global config
        :param modules: modules handle
        :param debug: debug flag (optional)
        :param rest_api: the rest api url
        """
        self.cfg = cfg
        self.modules = modules
        self.debug = debug
        self.statsd = statsd
        self.rest_api = rest_api
        self.current_user = None
        self.current_user_secure = None
        self.rest_client = None

    def set_default_headers(self):
        self._headers['Server'] = 'IceProd/' + iceprod.__version__

    def get_template_namespace(self):
        namespace = super(PublicHandler,self).get_template_namespace()
        namespace['version'] = iceprod.__version__
        namespace['section'] = self.request.uri.lstrip('/').split('?')[0].split('/')[0]
        namespace['master'] = ('master' in self.cfg and
                               'status' in self.cfg['master'] and
                               self.cfg['master']['status'])
        namespace['master_url'] = ('master' in self.cfg and
                                   'url' in self.cfg['master'] and
                                   self.cfg['master']['url'])
        namespace['site_id'] = (self.cfg['site_id'] if 'site_id' in self.cfg else None)
        namespace['sites'] = (self.cfg['webserver']['sites'] if (
                              'webserver' in self.cfg and
                              'sites' in self.cfg['webserver']) else None)
        namespace['json_encode'] = json_encode
        return namespace

    def prepare(self):
        try:
            data = self.get_secure_cookie("user", max_age_days=1)
            if not data:
                raise Exception('user cookie is missing/empty')
            data = json_decode(data)
            user_secure = self.get_secure_cookie("user_secure", max_age_days=0.01)
            if user_secure is not None and data['username'] != user_secure:
                raise Exception('mismatch between user_secure and username')
            self.current_user = data['username']
            self.current_user_data = data
            self.current_user_secure = (user_secure is not None)
            self.rest_client = RestClient(self.rest_api, data['token'], timeout=50)
        except Exception:
            logger.info('error getting current user', exc_info=True)
            self.clear_cookie("user")
            self.clear_cookie("user_secure")
            self.current_user = None

    def write_error(self,status_code=500,**kwargs):
        """Write out custom error page."""
        self.set_status(status_code)
        if status_code >= 500:
            self.write('<h2>Internal Error</h2>')
        else:
            self.write('<h2>Request Error</h2>')
        if 'message' in kwargs:
            self.write('<br />'.join(kwargs['message'].split('\n')))
        elif 'reason' in kwargs:
            self.write('<br />'.join(kwargs['reason'].split('\n')))
        elif self._reason:
            self.write('<br />'.join(self._reason.split('\n')))
        self.finish()

class Default(PublicHandler):
    """Handle / urls"""
    @catch_error
    async def get(self):
        self.statsd.incr('default')
        self.render('main.html')

class Submit(PublicHandler):
    """Handle /submit urls"""
    @catch_error
    @tornado.web.authenticated
    async def get(self):
        logger.info('here')
        self.statsd.incr('submit')
        url = self.request.uri[1:]
        ret = await self.rest_client.request('POST','/create_token')
        token = ret['result']
        groups = []
        logger.info('user_data: %r', self.current_user_data)
        logger.info('token: %r', token)
        if self.current_user_data and 'groups' in self.current_user_data:
            groups = self.current_user_data['groups']
        default_config = {
          "categories": [],
          "dataset": 0,
          "description": "",
          "difplus": None,
          "options": {
          },
          "parent_id": 0,
          "steering": None,
          "tasks": [],
          "version": 3
        }
        render_args = {
            'passkey':token,
            'edit':False,
            'dataset_id':'',
            'config':default_config,
            'groups':groups,
            'description':'',
        }
        self.render('submit.html',**render_args)

class Config(PublicHandler):
    """Handle /config urls"""
    @catch_error
    @tornado.web.authenticated
    async def get(self):
        self.statsd.incr('config')
        dataset_id = self.get_argument('dataset_id',default=None)
        if not dataset_id:
            self.write_error(400,message='must provide dataset_id')
            return
        dataset = await self.rest_client.request('GET','/datasets/{}'.format(dataset_id))
        edit = self.get_argument('edit',default=False)
        if edit:
            ret = await self.rest_client.request('POST','/create_token')
            passkey = ret['result']
        else:
            passkey = None
        config = await self.rest_client.request('GET','/config/{}'.format(dataset_id))
        render_args = {
            'edit':edit,
            'passkey':passkey,
            'dataset_id':dataset_id,
            'config':config,
            'description':dataset.get('description',''),
        }
        self.render('submit.html',**render_args)

class DatasetBrowse(PublicHandler):
    """Handle /dataset urls"""
    @catch_error
    @tornado.web.authenticated
    async def get(self):
        self.statsd.incr('dataset_browse')
        filter_options = {'status':['processing','suspended','errors','complete','truncated']}
        filter_results = {n:self.get_arguments(n) for n in filter_options}

        args = []
        for name in filter_results:
            val = filter_results[name]
            if any(v not in filter_options[name] for v in val):
                raise tornado.web.HTTPError(400, reason='Bad filter '+name+' value')
            args.append(name+'='+('|'.join(val)))

        url = '/datasets'
        if args:
            url += '?'+('&'.join(args))

        ret = await self.rest_client.request('GET', url)
        datasets = sorted(ret.values(), key=lambda x:x['dataset'], reverse=True)
        #logger.info('datasets: %r', datasets)
        self.render('dataset_browse.html',datasets=datasets,
                    filter_options=filter_options,
                    filter_results=filter_results)

class Dataset(PublicHandler):
    """Handle /dataset urls"""
    @catch_error
    @tornado.web.authenticated
    async def get(self, dataset_id):
        self.statsd.incr('dataset')

        if dataset_id.isdigit():
            try:
                d_num = int(dataset_id)
                if d_num < 10000000:
                    all_datasets = await self.rest_client.request('GET','/datasets')
                    for d in all_datasets.values():
                        if d['dataset'] == d_num:
                            dataset_id = d['dataset_id']
                            break
            except Exception:
                pass
        try:
            dataset = await self.rest_client.request('GET','/datasets/{}'.format(dataset_id))
        except Exception:
            raise tornado.web.HTTPError(404, reason='Dataset not found')
        dataset_num = dataset['dataset']

        ret = await self.rest_client.request('POST','/create_token')
        passkey = ret['result']

        jobs = await self.rest_client.request('GET','/datasets/{}/job_counts/status'.format(dataset_id))
        tasks = await self.rest_client.request('GET','/datasets/{}/task_counts/status'.format(dataset_id))
        task_info = await self.rest_client.request('GET','/datasets/{}/task_counts/name_status'.format(dataset_id))
        task_stats = await self.rest_client.request('GET','/datasets/{}/task_stats'.format(dataset_id))
        config = await self.rest_client.request('GET','/config/{}'.format(dataset_id))
        for t in task_info:
            logger.info('task_info[%s] = %r', t, task_info[t])
            for s in ('waiting','queued','processing','complete'):
                if s not in task_info[t]:
                    task_info[t][s] = 0
            error = 0
            for s in ('reset','resume','failed'):
                if s in task_info[t]:
                    error += task_info[t][s]
            task_info[t]['error'] = error
            for task in config['tasks']:
                if 'name' in task and task['name'] == t:
                    task_info[t]['type'] = 'GPU' if 'requirements' in task and 'gpu' in task['requirements'] and task['requirements']['gpu'] else 'CPU'
                    break
            else:
                task_info[t]['type'] = 'UNK'
        self.render('dataset_detail.html',dataset_id=dataset_id,dataset_num=dataset_num,
                    dataset=dataset,jobs=jobs,tasks=tasks,task_info=task_info,task_stats=task_stats,passkey=passkey)

class TaskBrowse(PublicHandler):
    """Handle /task urls"""
    @catch_error
    @tornado.web.authenticated
    async def get(self, dataset_id):
        self.statsd.incr('task_browse')
        status = self.get_argument('status',default=None)

        if status:
            tasks = await self.rest_client.request('GET','/datasets/{}/tasks?status={}'.format(dataset_id,status))
            for t in tasks:
                if 'job_index' not in tasks[t]:
                    job = await self.rest_client.request('GET', '/datasets/{}/jobs/{}'.format(dataset_id, tasks[t]['job_id']))
                    tasks[t]['job_index'] = job['job_index']
            ret = await self.rest_client.request('POST','/create_token')
            passkey = ret['result']
            self.render('task_browse.html',tasks=tasks, passkey=passkey)
        else:
            status = await self.rest_client.request('GET','/datasets/{}/task_counts/status'.format(dataset_id))
            self.render('tasks.html',status=status)

class Task(PublicHandler):
    """Handle /task urls"""
    @catch_error
    @tornado.web.authenticated
    async def get(self, dataset_id, task_id):
        self.statsd.incr('task')
        status = self.get_argument('status',default=None)

        ret = await self.rest_client.request('POST','/create_token')
        passkey = ret['result']

        dataset = await self.rest_client.request('GET', '/datasets/{}'.format(dataset_id))
        task_details = await self.rest_client.request('GET','/datasets/{}/tasks/{}'.format(dataset_id, task_id))
        task_stats = await self.rest_client.request('GET','/datasets/{}/tasks/{}/task_stats?last=true'.format(dataset_id, task_id))
        if task_stats:
            task_stats = list(task_stats.values())[0]
        try:
            ret = await self.rest_client.request('GET','/datasets/{}/tasks/{}/logs?group=true'.format(dataset_id, task_id))
            logs = ret['logs']
            #logger.info("logs: %r", logs)
            ret2 = await self.rest_client.request('GET','/datasets/{}/tasks/{}/logs?keys=log_id|name|timestamp'.format(dataset_id, task_id))
            logs2 = ret2['logs']
            logger.info("logs2: %r", logs2)
            log_by_name = defaultdict(list)
            for log in sorted(logs2,key=lambda l:l['timestamp'] if 'timestamp' in l else '',reverse=True):
                log_by_name[log['name']].append(log)
            for log in logs:
                if 'data' not in log or not log['data']:
                    log['data'] = ''
                log_by_name[log['name']][0] = log
        except Exception:
            log_by_name = {}
        self.render('task_detail.html', dataset=dataset, task=task_details, task_stats=task_stats, logs=log_by_name, passkey=passkey)

class JobBrowse(PublicHandler):
    """Handle /job urls"""
    @catch_error
    @tornado.web.authenticated
    async def get(self, dataset_id):
        self.statsd.incr('job')
        status = self.get_argument('status',default=None)

        ret = await self.rest_client.request('POST','/create_token')
        passkey = ret['result']

        jobs = await self.rest_client.request('GET', '/datasets/{}/jobs'.format(dataset_id))
        if status:
            for t in list(jobs):
                if jobs[t]['status'] != status:
                    del jobs[t]
                    continue
        self.render('job_browse.html', jobs=jobs, passkey=passkey)

class Job(PublicHandler):
    """Handle /job urls"""
    @catch_error
    @tornado.web.authenticated
    async def get(self, dataset_id, job_id):
        self.statsd.incr('job')
        status = self.get_argument('status',default=None)

        ret = await self.rest_client.request('POST','/create_token')
        passkey = ret['result']

        dataset = await self.rest_client.request('GET', '/datasets/{}'.format(dataset_id))
        job = await self.rest_client.request('GET', '/datasets/{}/jobs/{}'.format(dataset_id,job_id))
        tasks = await self.rest_client.request('GET','/datasets/{}/tasks?job_id={}'.format(dataset_id,job_id))
        job['tasks'] = list(tasks.values())
        job['tasks'].sort(key=lambda x:x['task_index'])
        self.render('job_detail.html', dataset=dataset, job=job, passkey=passkey)

class Documentation(PublicHandler):
    @catch_error
    def get(self, url):
        self.statsd.incr('documentation')
        doc_path = get_pkgdata_filename('iceprod.server','data/docs')
        self.write(documentation.load_doc(doc_path+'/' + url))
        self.flush()

class Log(PublicHandler):
    @catch_error
    @tornado.web.authenticated
    async def get(self, dataset_id, log_id):
        self.statsd.incr('log')
        ret = await self.rest_client.request('GET','/datasets/{}/logs/{}'.format(dataset_id, log_id))
        log_text = ret['data']
        html = '<html><head><title>' + ret['name'] + '</title></head><body>'
        html += log_text.replace('\n', '<br/>')
        html += '</body></html>'
        self.write(html)
        self.flush()

class Help(PublicHandler):
    """Help Page"""
    @catch_error
    def get(self):
        self.statsd.incr('help')
        self.render('help.html')

class Other(PublicHandler):
    """Handle any other urls - this is basically all 404"""
    @catch_error
    def get(self):
        self.statsd.incr('other')
        path = self.request.path
        self.set_status(404)
        self.render('404.html',path=path)

class Profile(PublicHandler):
    """Handle user profile page"""
    @catch_error
    @tornado.web.authenticated
    async def get(self):
        self.statsd.incr('profile')
        ret = await self.rest_client.request('POST','/create_token')
        token = ret['result']
        groups = []
        logger.info('user_data: %r', self.current_user_data)
        logger.info('token: %r', token)
        if self.current_user_data and 'groups' in self.current_user_data:
            groups = self.current_user_data['groups']
        self.render('profile.html', username=self.current_user, groups=groups,
                    token=token)

class Login(PublicHandler):
    """Handle the login url"""
    def initialize(self, module_rest_client, *args, **kwargs):
        """
        Get some params from the website module

        :param module_rest_client: a REST Client
        """
        super(Login, self).initialize(*args, **kwargs)
        self.module_rest_client = module_rest_client

    @catch_error
    def get(self):
        self.statsd.incr('login')
        n = self.get_argument('next', default='/')
        secure = self.get_argument('secure', default=None)
        self.clear_cookie("user")
        self.clear_cookie("user_secure")
        self.render('login.html', status=None, next=n)

    @catch_error
    async def post(self):
        n = self.get_argument('next', default='/')
        secure = self.get_argument('secure', default=None)
        username = self.get_argument('username')
        password = self.get_argument('password')
        self.clear_cookie("user")
        self.clear_cookie("user_secure")
        try:
            data = await self.module_rest_client.request('POST','/ldap',{'username':username,'password':password})
            cookie = json_encode(data)
            if secure:
                self.set_secure_cookie('user_secure', username, expires_days=0.01)
            self.set_secure_cookie('user', cookie, expires_days=1)
            self.redirect(n)
        except Exception:
            logger.info('failed', exc_info=True)
            self.render('login.html', status='failed', next=n)

class Logout(PublicHandler):
    @catch_error
    def get(self):
        self.statsd.incr('logout')
        self.clear_cookie("user")
        self.clear_cookie("user_secure")
        self.current_user = None
        self.render('logout.html', status=None)



"""
A supercomputer plugin, specificially for
`Graham <https://docs.computecanada.ca/wiki/Graham>`_.
"""
from __future__ import print_function
import os
import sys
import stat
import random
import math
import logging
import getpass
import socket
from datetime import datetime,timedelta
import subprocess
import asyncio
from functools import partial

import tornado.gen
from tornado.concurrent import run_on_executor

import iceprod
from iceprod.core import dataclasses
from iceprod.core import constants
from iceprod.core import functions
from iceprod.core.exe_json import ServerComms
from iceprod.server import grid
from iceprod.server.globus import SiteGlobusProxy

logger = logging.getLogger('plugin-graham')

async def subprocess_ssh(host, args):
    """A subprocess call over ssh"""
    cmd = ['ssh', '-i', '~/.ssh/iceprod', host]
    if 'ICEPRODBASE' in os.environ:
        cmd.append(f'{os.environ["ICEPRODBASE"]}/env-shell.sh')
    cmd += args
    logger.info('subprocess_ssh: %r', cmd)
    p = await asyncio.create_subprocess_exec(*cmd)
    await asyncio.wait_for(p.wait(), timeout=20*60)
    if p.returncode:
        raise Exception(f'command failed, return code {p.returncode}')
    return p

async def check_call(*args, **kwargs):
    logger.info('subprocess_check_call: %r', args)
    p = await asyncio.create_subprocess_exec(*args, **kwargs)
    if p.returncode:
        raise Exception(f'command failed, return code {p.returncode}')
    return p

async def check_output(*args, **kwargs):
    kwargs['stdout'] = subprocess.PIPE
    kwargs['stderr'] = subprocess.STDOUT
    logger.info('subprocess_check_output: %r', args)
    p = await asyncio.create_subprocess_exec(*args, **kwargs)
    out,_ = await p.communicate()
    if p.returncode:
        raise Exception(f'command failed, return code {p.returncode}')
    return out.decode('utf-8')

class MyServerComms(ServerComms):
    def __init__(self, rest_client):
        self.rest = rest_client

class supercomp_graham(grid.BaseGrid):

    ### Plugin Overrides ###

    # graham gpu queue requirements
    resources = {
        'cpu': 2,
        'memory': 8.,
        'disk': 48.,
        'gpu': 1,
        'time': 4.,
    }

    def __init__(self, *args, **kwargs):
        super(supercomp_graham, self).__init__(*args, **kwargs)
        self.x509proxy = SiteGlobusProxy()

    async def upload_logfiles(self, task_id, dataset_id, submit_dir='', reason=''):
        """upload logfiles"""
        data = {'name': 'stdlog', 'task_id': task_id, 'dataset_id': dataset_id}

        # upload stdlog
        filename = os.path.join(submit_dir, constants['stdlog'])
        if os.path.exists(filename):
            with open(filename) as f:
                data['data'] = f.read()
            await self.rest_client.request('POST', '/logs', data)
        else:
            data['data'] = reason
            await self.rest_client.request('POST', '/logs', data)

        # upload stderr
        data['name'] = 'stderr'
        filename = os.path.join(submit_dir, constants['stderr'])
        if os.path.exists(filename):
            with open(filename) as f:
                data['data'] = f.read()
            await self.rest_client.request('POST', '/logs', data)
        else:
            data['data'] = ''
            await self.rest_client.request('POST', '/logs', data)

        # upload stdout
        data['name'] = 'stdout'
        filename = os.path.join(submit_dir, constants['stdout'])
        if os.path.exists(filename):
            with open(filename) as f:
                data['data'] = f.read()
            await self.rest_client.request('POST', '/logs', data)
        else:
            data['data'] = ''
            await self.rest_client.request('POST', '/logs', data)

    async def task_error(self, task_id, dataset_id, submit_dir, reason=''):
        """reset a task"""
        # search for resources in slurm stdout
        resources = {}
        filename = os.path.join(submit_dir, 'slurm.out')
        if os.path.exists(filename):
            with open(filename) as f:
                resource_lines = False
                for line in f:
                    line = line.strip()
                    if line == 'Resources:':
                        resource_lines = True
                        continue
                    if resource_lines:
                        if ':' in line:
                            name,value = line.split(':',1)
                            if value.isdigit():
                                resources[name] = int(value)
                            else:
                                try:
                                    resources[name] = float(value)
                                except Exception:
                                    resources[name] = value
                            continue
                        else:
                            break
        if not reason:
            # search for reason in logfile
            filename = os.path.join(submit_dir, constants['stdlog'])
            if os.path.exists(filename):
                with open(filename) as f:
                    for line in f:
                        line = line.strip()
                        if 'failed to download' in line:
                            reason = 'failed to download input file(s)'
                        if 'failed to upload' in line:
                            reason = 'failed to upload output file(s)'
                        if 'Exception' in line and not reason:
                            reason = line
                        if 'return code:' in line and not reason:
                            reason = 'task error: return code '+line.rsplit(':',1)[-1]

        comms = MyServerComms(self.rest_client)
        await comms.task_kill(task_id, dataset_id=dataset_id, 
                              reason=reason, resources=resources)

    async def finish_task(self, task_id, dataset_id, submit_dir):
        """complete a task"""
        # search for reasources in slurm stdout
        resources = {}
        filename = os.path.join(submit_dir, 'slurm.out')
        if os.path.exists(filename):
            with open(filename) as f:
                resource_lines = False
                for line in f:
                    line = line.strip()
                    if line == 'Resources:':
                        resource_lines = True
                        continue
                    if resource_lines:
                        if ':' in line:
                            name,value = line.split(':',1)
                            if value.isdigit():
                                resources[name] = int(value)
                            else:
                                try:
                                    resources[name] = float(value)
                                except Exception:
                                    resources[name] = value
                            continue
                        else:
                            break

        comms = MyServerComms(self.rest_client)
        await comms.finish_task(task_id, dataset_id=dataset_id, 
                                resources=resources)

    async def check_and_clean(self):
        """Check and clean the grid"""
        host = socket.getfqdn()
        self.x509proxy.update_proxy()

        # first, check if anything has completed successfully
        grid_jobs = await self.get_grid_completions()

        if grid_jobs:
            # get pilots from iceprod
            pilots = await self.rest_client.request('GET', '/pilots')
            pilot_futures = []
            pilots_to_delete = set()
            for pilot_id in pilots:
                pilot = pilots[pilot_id]
                if pilot['queue_host'] != host:
                    continue
                if 'grid_queue_id' in pilot and pilot['grid_queue_id'] in grid_jobs:
                    try:
                        gid = pilot['grid_queue_id']
                        pilot['submit_dir'] = grid_jobs[gid]['submit_dir']
                        if pilot['tasks']:
                            task_id = pilot['tasks'][0]
                            ret = await self.rest_client.request('GET', f'/tasks/{task_id}')
                            if ret['status'] == 'processing':
                                pilot['dataset_id'] = ret['dataset_id']
                                if grid_jobs[gid]['status'] == 'ok':
                                    # upload data - do other steps in next loop
                                    pilot_futures.append(asyncio.ensure_future(self.upload_output(pilot)))
                                else:
                                    await self.upload_logfiles(task_id, pilot['dataset_id'],
                                                               submit_dir=pilot['submit_dir'])
                                    await self.task_error(task_id, pilot['dataset_id'],
                                                          submit_dir=pilot['submit_dir'])
                    except Exception:
                        logger.error('error handling task', exc_info=True)

                    pilots_to_delete.add(pilot_id)

            for fut in asyncio.as_completed(pilot_futures):
                pilot,e = await fut # upload is done

                try:
                    task_id = pilot['tasks'][0]
                    if e:
                        reason = f'failed to upload output files\n{e}'
                        await self.upload_logfiles(task_id,
                                                   dataset_id=pilot['dataset_id'],
                                                   submit_dir=pilot['submit_dir'],
                                                   reason=reason)
                        await self.task_error(task_id,
                                              dataset_id=pilot['dataset_id'],
                                              submit_dir=pilot['submit_dir'],
                                              reason=reason)
                    else:
                        await self.upload_logfiles(task_id,
                                                   dataset_id=pilot['dataset_id'],
                                                   submit_dir=pilot['submit_dir'])
                        await self.finish_task(task_id,
                                               dataset_id=pilot['dataset_id'],
                                               submit_dir=pilot['submit_dir'])
                except Exception:
                    logger.error('error handling task', exc_info=True)

            for pilot_id in pilots_to_delete:
                await self.rest_client.request('DELETE', f'/pilots/{pilot_id}')

        # now, do regular check_and_clean
        await super(supercomp_graham, self).check_and_clean()

    async def queue(self):
        """Submit a pilot for each task, up to the limit"""
        host = socket.getfqdn()
        self.x509proxy.update_proxy()
        resources = self.resources.copy()

        debug = False
        if ('queue' in self.cfg and 'debug' in self.cfg['queue']
            and self.cfg['queue']['debug']):
            debug = True

        dataset_cache = {}
        task_futures = []
        for _ in range(self.get_queue_num()):
            # get a processing task
            args = {'requirements': supercomp_graham.resources.copy()}
            args['requirements']['os'] = 'RHEL_7_x86_64'
            try:
                task = await self.rest_client.request('POST', f'/task_actions/process', args)
            except Exception:
                logger.info('no more tasks to queue')
                break

            # get full job, dataset, config info
            job = await self.rest_client.request('GET', f'/jobs/{task["job_id"]}')
            if task['dataset_id'] in dataset_cache:
                dataset, config = dataset_cache[task['dataset_id']]
            else:
                dataset = await self.rest_client.request('GET', f'/datasets/{task["dataset_id"]}')
                config = await self.rest_client.request('GET', f'/config/{task["dataset_id"]}')
                dataset_cache[task['dataset_id']] = (dataset, config)

            # check if we have any files in the task_files API
            task_cfg = None
            for t in config['tasks']:
                if t['name'] == task['name']:
                    task_cfg = t
                    break
            else:
                logger.warning('cannot find task in config for %s', task['task_id'])
                continue
            if 'task_files' in task_cfg and task_cfg['task_files']:
                comms = MyServerComms(self.rest_client)
                files = await comms.task_files(task['dataset_id'],
                                               task['task_id'])
                task_cfg['data'].extend(files)

            config['dataset'] = dataset['dataset']
            task.update({
                'config': config,
                'job': job['job_index'],
                'dataset': dataset['dataset'],
                'jobs_submitted': dataset['jobs_submitted'],
                'tasks_submitted': dataset['tasks_submitted'],
                'debug': dataset['debug'],
                'requirements': args['requirements'],
            })

            # setup submit dir
            await self.setup_submit_directory(task)

            # create pilot
            resources_available = {'time': resources['time']}
            for k in ('cpu','gpu','memory','disk'):
                resources_available[k] = resources[k]-task['requirements'][k]
            pilot = {'resources': resources,
                     'resources_available': resources_available,
                     'resources_claimed': task['requirements'],
                     'tasks': [task['task_id']],
                     'queue_host': host,
                     'queue_version': iceprod.__version__,
                     'version': iceprod.__version__,
            }
            ret = await self.rest_client.request('POST', '/pilots', pilot)
            pilot['pilot_id'] = ret['result']
            task['pilot'] = pilot

            # get input files, all tasks in parallel
            task_futures.append(asyncio.ensure_future(self.download_input(task)))

        # wait for the futures
        for fut in asyncio.as_completed(task_futures):
            task,e = await fut
            try:
                pilot_id = task['pilot']['pilot_id']
                if e:
                    reason = f'failed to download input files\n{e}'
                    await self.upload_logfiles(task['task_id'],
                                               dataset_id=task['dataset_id'],
                                               submit_dir=task['submit_dir'],
                                               reason=reason)
                    await self.task_error(task['task_id'],
                                          dataset_id=task['dataset_id'],
                                          submit_dir=task['submit_dir'],
                                          reason=reason)
                    await self.rest_client.request('DELETE', f'/pilots/{pilot_id}')
                    continue

                # submit to queue
                await self.submit(task)

                # update pilot
                args = {'grid_queue_id': task['grid_queue_id']}
                await self.rest_client.request('PATCH', f'/pilots/{pilot_id}', args)
            except Exception as e:
                try:
                    reason = f'failed to submit pilot:\n{e}'
                    await self.upload_logfiles(task['task_id'],
                                               dataset_id=task['dataset_id'],
                                               submit_dir=task['submit_dir'],
                                               reason=reason)
                    await self.task_error(task['task_id'],
                                          dataset_id=task['dataset_id'],
                                          submit_dir=task['submit_dir'],
                                          reason=reason)
                except Exception:
                    pass
                logger.error('error handling pilot', exc_info=True)

    async def download_input(self, task):
        """
        Download input files for task.

        Args:
            task (dict): task info
        Returns:
            dict: task info
        """
        try:
            #proxy = self.x509proxy.get_proxy()
            await subprocess_ssh(
                    'gra-dtn1.computecanada.ca',
                    ['python', '-m', 'iceprod.core.data_transfer', '-f',
                     os.path.join(task['submit_dir'],'task.cfg'),
                     '-d', task['submit_dir'],
                     'input']
            )
        except Exception as e:
            return (task, e)
        return (task,None)

    async def upload_output(self, task):
        """
        Upload output files for task.

        Args:
            task (dict): task info
        Returns:
            dict: task info
        """
        try:
            #proxy = self.x509proxy.get_proxy()
            await subprocess_ssh(
                    'gra-dtn1.computecanada.ca',
                    ['python', '-m', 'iceprod.core.data_transfer', '-f',
                     os.path.join(task['submit_dir'],'task.cfg'),
                     '-d', task['submit_dir'],
                     'output']
            )
        except Exception as e:
            return (task,e)
        return (task,None)

    @run_on_executor
    def generate_submit_file(self, task, cfg=None, passkey=None,
                             filelist=None):
        """Generate queueing system submit file for task in dir."""
        args = self.get_submit_args(task,cfg=cfg)
        args.append('--offline')

        # write the submit file
        submit_file = os.path.join(task['submit_dir'],'submit.sh')
        with open(submit_file,'w') as f:
            p = partial(print,sep='',file=f)
            p('#!/bin/bash')
            p('#SBATCH --account=def-dgrant')
            p('#SBATCH --output={}'.format(os.path.join(task['submit_dir'],'slurm.out')))
            p('#SBATCH --error={}'.format(os.path.join(task['submit_dir'],'slurm.err')))
            p(f'#SBATCH --chdir={task["submit_dir"]}')
            p('#SBATCH --ntasks=1')
            p('#SBATCH --export=NONE')
            p('#SBATCH --mail-type=NONE')
            p('#SBATCH --job-name=iceprod_{}'.format(os.path.basename(task['submit_dir'])))

            # handle resource requests
            if 'requirements' in task:
                if 'cpu' in task['requirements'] and task['requirements']['cpu']:
                    p(f'#SBATCH --cpus-per-task={task["requirements"]["cpu"]}')
                if 'gpu' in task['requirements'] and task['requirements']['gpu']:
                    p(f'#SBATCH --gres=gpu:{task["requirements"]["gpu"]}')
                if 'memory' in task['requirements'] and task['requirements']['memory']:
                    p('#SBATCH --mem={}M'.format(int(task['requirements']['memory']*1000)))
                # we don't currently use the local disk, just the global scratch
                #if 'disk' in task['requirements'] and task['requirements']['disk']:
                #    p('#SBATCH --tmp={}M'.format(int(task['requirements']['disk']*1000)))
                if 'time' in task['requirements'] and task['requirements']['time']:
                    p('#SBATCH --time={}'.format(int(task['requirements']['time']*60)))

            # get batchopts
            for b in self.queue_cfg['batchopts']:
                p(b+'='+self.queue_cfg['batchopts'][b])

            # make resources explicit in env
            if 'requirements' in task:
                if 'cpu' in task['requirements'] and task['requirements']['cpu']:
                    p(f'export NUM_CPUS={task["requirements"]["cpu"]}')
                if 'gpu' in task['requirements'] and task['requirements']['gpu']:
                    p(f'export NUM_GPUS={task["requirements"]["gpu"]}')
                if 'memory' in task['requirements'] and task['requirements']['memory']:
                    p(f'export NUM_MEMORY={task["requirements"]["memory"]}')
                if 'disk' in task['requirements'] and task['requirements']['disk']:
                    p(f'export NUM_DISK={task["requirements"]["disk"]}')
                if 'time' in task['requirements'] and task['requirements']['time']:
                    p(f'export NUM_TIME={task["requirements"]["time"]}')

            p('{} {}'.format(os.path.join(task['submit_dir'],'loader.sh'), ' '.join(args)))

        # make it executable
        st = os.stat(submit_file)
        os.chmod(submit_file, st.st_mode | stat.S_IEXEC)

    async def submit(self,task):
        """Submit task to queueing system."""
        cmd = ['sbatch','submit.sh']
        ret = await check_output(*cmd, cwd=task['submit_dir'])
        grid_queue_id = ''
        for line in ret.split('\n'):
            if 'Submitted batch job' in line:
                grid_queue_id = line.strip().rsplit(' ',1)[-1]
                break
        else:
            raise Exception('did not get a grid_queue_id')
        task['grid_queue_id'] = grid_queue_id

    async def get_grid_status(self):
        """
        Get all tasks running on the queue system.
        
        Returns:
            dict: {grid_queue_id: {status, submit_dir} }
        """
        cmd = ['squeue', '-u', getpass.getuser(), '-h', '-o', '%A %t %j %o']
        out = await check_output(*cmd)
        ret = {}
        for line in out.split('\n'):
            if not line.strip():
                continue
            gid,status,name,cmd = line.split()
            if not name.startswith('iceprod'):
                continue
            if status == 'PD':
                status = 'queued'
            elif status == 'R':
                status = 'processing'
            elif status == 'CD':
                status = 'completed'
            else:
                status = 'error'
            ret[gid] = {'status':status,'submit_dir':os.path.dirname(cmd)}
        return ret

    async def get_grid_completions(self):
        """
        Get completions in the last 4 days.
        
        Returns:
            dict: {grid_queue_id: {status, submit_dir} }
        """
        date = (datetime.now()-timedelta(days=4)).isoformat().split('.',1)[0]
        cmd = ['sacct', '-u', getpass.getuser(), '-n', '-P', '-S', date, '-o', 'JobIDRaw,State,JobName,ExitCode,Workdir']
        out = await check_output(*cmd)
        ret = {}
        for line in out.split('\n'):
            if not line.strip():
                continue
            gid,status,name,exit_code,workdir = line.strip().split('|')
            if status != 'COMPLETED' or not name.startswith('iceprod'):
                continue
            if exit_code == '0:0':
                status = 'ok'
            else:
                status = 'error'
            ret[gid] = {'status':status,'submit_dir':workdir}
        return ret

    async def remove(self,tasks):
        """Remove tasks from queueing system."""
        if tasks:
            cmd = ['scancel']+list(tasks)
            await check_call(*cmd)



"""
Clean job temp directories.

Check job temp directories, and if the job is complete then delete it.

Initial delay: rand(15 minutes)
Periodic delay: 60 minutes
"""

import os
import logging
import random
import time
from datetime import datetime, timedelta
from functools import partial
from asyncio import wrap_future

from tornado.ioloop import IOLoop

from iceprod.core.gridftp import GridFTP
from iceprod.server.util import str2datetime

logger = logging.getLogger('job_temp_cleaning')

def job_temp_cleaning(module):
    """
    Initial entrypoint.

    Args:
        module (:py:class:`iceprod.server.modules.schedule`): schedule module
    """
    # initial delay
    IOLoop.current().call_later(random.randint(10,60*10), run, module.rest_client,
            module.cfg, module.executor)

async def run(rest_client, cfg, executor, dataset=None, debug=False):
    """
    Actual runtime / loop.

    Args:
        rest_client (:py:class:`iceprod.core.rest_client.Client`): rest client
        cfg (:py:class:`iceprod.server.config.IceProdConfig`): iceprod config
        debug (bool): debug flag to propagate exceptions
    """
    if 'queue' not in cfg or 'site_temp' not in cfg['queue']:
        return

    start_time = time.time()

    suspend_time = timedelta(days=90)
    now = datetime.utcnow()

    try:
        # get all the job_indexes currently in tmp
        temp_dir = cfg['queue']['site_temp']
        dataset_dirs = await wrap_future(executor.submit(partial(GridFTP.list, temp_dir, details=True)))
        logger.debug('dataset_dirs: %r', dataset_dirs)
        ret = await rest_client.request('GET', '/datasets?keys=dataset_id|dataset')
        datasets = {}
        for d in ret:
            datasets[str(ret[d]['dataset'])] = ret[d]['dataset_id']
        for entry in dataset_dirs:
            if not entry.directory:
                continue
            d = entry.name
            if dataset and d != dataset:
                continue
            logger.info('temp cleaning for dataset %r', d)
            try:
                job_dirs = await wrap_future(executor.submit(partial(GridFTP.list, os.path.join(temp_dir, d), details=True)))
            except Exception:
                logger.error('failed to get job dirs for dataset %r', d, exc_info=True)
                if debug:
                    raise
                continue
            logger.debug('job_dirs: %r', job_dirs)
            try:
                jobs = await rest_client.request('GET', f'/datasets/{datasets[d]}/jobs', {'keys':'status|status_changed|job_index'})
            except Exception:
                logger.error('failed to get jobs for dataset %r', d, exc_info=True)
                if debug:
                    raise
                continue
            logger.debug('jobs: %r', jobs)
            job_indexes = set()
            for job in jobs.values():
                if job['status'] == 'complete' or (job['status'] != 'processing'
                    and now - str2datetime(job['status_changed']) > suspend_time):
                    job_indexes.add(job['job_index'])
            logger.debug('job_indexes: %r', job_indexes)
            for job in job_dirs:
                if not job.directory:
                    continue
                j = job.name
                if not j.isnumeric():
                    if debug:
                        logger.info('j is not numeric: %r', j)
                        raise Exception('not numeric')
                    continue
                if int(j) in job_indexes:
                    try:
                        dagtemp = os.path.join(temp_dir, d, j)
                        logger.info('cleaning site_temp %r', dagtemp)
                        await wrap_future(executor.submit(partial(GridFTP.rmtree, dagtemp)))
                    except Exception:
                        logger.warning('failed to clean site_temp', exc_info=True)
                        if debug:
                            raise

    except Exception:
        logger.error('error checking job temp', exc_info=True)
        if debug:
            raise

    # run again after 60 minute delay
    stop_time = time.time()
    delay = max(60*60 - (stop_time-start_time), 60*10)
    IOLoop.current().call_later(delay, run, rest_client, cfg, executor, debug)


def main():
    import argparse
    parser = argparse.ArgumentParser(description='run a scheduled task once')
    parser.add_argument('-t', '--token', default=os.environ.get('ICEPROD_TOKEN', None), help='auth token')
    parser.add_argument('-d', '--dataset', type=str, help='dataset num (optional)')
    parser.add_argument('--site-temp', default='gsiftp://gridftp-scratch.icecube.wisc.edu/mnt/tank/simprod', help='site temp location')
    parser.add_argument('--log-level', default='info', help='log level')
    parser.add_argument('--debug', default=False, action='store_true', help='debug enabled')

    args = parser.parse_args()
    args = vars(args)

    logformat='%(asctime)s %(levelname)s %(name)s %(module)s:%(lineno)s - %(message)s'
    logging.basicConfig(format=logformat, level=getattr(logging, args['log_level'].upper()))

    from rest_tools.client import RestClient
    rpc = RestClient('https://iceprod2-api.icecube.wisc.edu', args['token'])

    from iceprod.server.config import IceProdConfig
    cfg = IceProdConfig()
    if 'queue' not in cfg:
        cfg['queue'] = {}
    if 'site_temp' not in cfg['queue']:
        cfg['queue']['site_temp'] = args['site_temp']

    from concurrent.futures import ThreadPoolExecutor
    pool = ThreadPoolExecutor()

    import asyncio
    asyncio.run(run(rpc, cfg, pool, dataset=args['dataset'], debug=args['debug']))

if __name__ == '__main__':
    main()



import logging
import json
import uuid

import tornado.web
import pymongo
import motor

from iceprod.server.rest import RESTHandler, RESTHandlerSetup, authorization
from iceprod.server.util import nowstr

logger = logging.getLogger('rest.grids')

def setup(config, *args, **kwargs):
    """
    Setup method for Grids REST API.

    Sets up any database connections or other prerequisites.

    Args:
        config (dict): an instance of :py:class:`iceprod.server.config`.

    Returns:
        list: Routes for logs, which can be passed to :py:class:`tornado.web.Application`.
    """
    cfg_rest = config.get('rest',{}).get('grids',{})
    db_cfg = cfg_rest.get('database',{})

    # add indexes
    db = pymongo.MongoClient(**db_cfg).grids
    if 'grid_id_index' not in db.grids.index_information():
        db.grids.create_index('grid_id', name='grid_id_index', unique=True)

    handler_cfg = RESTHandlerSetup(config, *args, **kwargs)
    handler_cfg.update({
        'database': motor.motor_tornado.MotorClient(**db_cfg).grids,
    })

    return [
        (r'/grids', MultiGridsHandler, handler_cfg),
        (r'/grids/(?P<grid_id>\w+)', GridsHandler, handler_cfg),
    ]

class BaseHandler(RESTHandler):
    def initialize(self, database=None, **kwargs):
        super(BaseHandler, self).initialize(**kwargs)
        self.db = database

class MultiGridsHandler(BaseHandler):
    """
    Handle multi grids requests.
    """
    @authorization(roles=['admin','client'])
    async def get(self):
        """
        Get grid entries.

        Returns:
            dict: {'uuid': {grid_data}}
        """
        ret = await self.db.grids.find(projection={'_id':False}).to_list(1000)
        self.write({row['grid_id']:row for row in ret})
        self.finish()

    @authorization(roles=['admin','client'])
    async def post(self):
        """
        Create a grid entry.

        Body should contain the grid data.

        Returns:
            dict: {'result': <grid_id>}
        """
        data = json.loads(self.request.body)

        # validate first
        req_fields = {
            'host': str,
            'queues': dict, # dict of {name: type}
            'version': str, # iceprod version
        }
        for k in req_fields:
            if k not in data:
                raise tornado.web.HTTPError(400, reason='missing key: '+k)
            if not isinstance(data[k], req_fields[k]):
                r = 'key {} should be of type {}'.format(k, req_fields[k])
                raise tornado.web.HTTPError(400, reason=r)

        # set some fields
        grid_id = uuid.uuid1().hex
        data['grid_id'] = grid_id
        data['start_date'] = nowstr()
        data['last_update'] = data['start_date']
        data['username'] = self.auth_data['username']
        if 'debug' not in data:
            data['debug'] = False

        ret = await self.db.grids.insert_one(data)
        self.set_status(201)
        self.write({'result': grid_id})
        self.finish()

class GridsHandler(BaseHandler):
    """
    Handle single grid requests.
    """
    @authorization(roles=['admin','client'])
    async def get(self, grid_id):
        """
        Get a grid entry.

        Args:
            grid_id (str): the grid id

        Returns:
            dict: grid entry
        """
        ret = await self.db.grids.find_one({'grid_id':grid_id},
                projection={'_id':False})
        if not ret:
            self.send_error(404, reason="Grid not found")
        else:
            self.write(ret)
            self.finish()

    @authorization(roles=['admin','client'])
    async def patch(self, grid_id):
        """
        Update a grid entry.

        Body should contain the grid data to update.  Note that this will
        perform a merge (not replace).

        Args:
            grid_id (str): the grid id

        Returns:
            dict: updated grid entry
        """
        data = json.loads(self.request.body)
        if not data:
            raise tornado.web.HTTPError(400, reason='Missing update data')
        data['last_update'] = nowstr()

        ret = await self.db.grids.find_one_and_update({'grid_id':grid_id},
                {'$set':data},
                projection={'_id':False},
                return_document=pymongo.ReturnDocument.AFTER)
        if not ret:
            self.send_error(404, reason="Grid not found")
        else:
            self.write(ret)
            self.finish()



import logging
import json
import uuid
import io
from functools import partial
import asyncio
from concurrent.futures import ThreadPoolExecutor

import tornado.web
import pymongo
import motor

try:
    import boto3
    import botocore.client
    import botocore.exceptions
except ImportError:
    boto3 = None

from iceprod.server.util import nowstr
from iceprod.server.rest import RESTHandler, RESTHandlerSetup, authorization, catch_error

logger = logging.getLogger('rest.logs')


class S3:
    """S3 wrapper for uploading and downloading objects"""
    def __init__(self, config):
        self.s3 = None
        self.bucket = 'iceprod2-logs'
        if (boto3 and 's3' in config and 'access_key' in config['s3'] and
            'secret_key' in config['s3']):
            try:
                self.s3 = boto3.client('s3','us-east-1',
                    endpoint_url=config['s3'].get('host', None),
                    aws_access_key_id=config['s3']['access_key'],
                    aws_secret_access_key=config['s3']['secret_key'],
                    config=botocore.client.Config(max_pool_connections=101))
            except Exception:
                logger.warning('failed to connect to s3: %r',
                            config['s3'], exc_info=True)
        self.executor = ThreadPoolExecutor(max_workers=20)

    async def get(self, key):
        """Download object from S3"""
        ret = ''
        with io.BytesIO() as f:
            loop = asyncio.get_event_loop()
            try:
                await loop.run_in_executor(self.executor,
                        partial(self.s3.download_fileobj, Bucket=self.bucket,
                                Key=key, Fileobj=f))
                ret = f.getvalue()
            except botocore.exceptions.ClientError as e:
                error_code = e.response['Error']['Code']
                if error_code == '404':
                    return '' # don't error on a 404
                raise
        return ret.decode('utf-8')

    async def put(self, key, data):
        """Upload object to S3"""
        with io.BytesIO(data.encode('utf-8')) as f:
            loop = asyncio.get_event_loop()
            await loop.run_in_executor(self.executor,
                    partial(self.s3.upload_fileobj, f, self.bucket, key))

    async def exists(self, key):
        """Check existence in S3"""
        loop = asyncio.get_event_loop()
        try:
            await loop.run_in_executor(self.executor,
                    partial(self.s3.head_object, Bucket=self.bucket, Key=key))
        except Exception:
            return False
        return True

    async def delete(self, key):
        """Delete object in S3"""
        loop = asyncio.get_event_loop()
        await loop.run_in_executor(self.executor,
                partial(self.s3.delete_object, Bucket=self.bucket, Key=key))


def setup(config, *args, **kwargs):
    """
    Setup method for Logs REST API.

    Sets up any database connections or other prerequisites.

    Args:
        config (dict): an instance of :py:class:`iceprod.server.config`.

    Returns:
        list: Routes for logs, which can be passed to :py:class:`tornado.web.Application`.
    """
    cfg_rest = config.get('rest',{}).get('logs',{})
    db_cfg = cfg_rest.get('database',{})

    # add indexes
    db = pymongo.MongoClient(**db_cfg).logs
    if 'log_id_index' not in db.logs.index_information():
        db.logs.create_index('log_id', name='log_id_index', unique=True)
    if 'task_id_index' not in db.logs.index_information():
        db.logs.create_index('task_id', name='task_id_index', unique=False)
    if 'dataset_id_index' not in db.logs.index_information():
        db.logs.create_index('dataset_id', name='dataset_id_index', unique=False)

    # S3
    s3 = S3(config) if boto3 and 's3' in config else None

    handler_cfg = RESTHandlerSetup(config, *args, **kwargs)
    handler_cfg.update({
        'database': motor.motor_tornado.MotorClient(**db_cfg).logs,
        's3': s3,
    })

    return [
        (r'/logs', MultiLogsHandler, handler_cfg),
        (r'/logs/(?P<log_id>\w+)', LogsHandler, handler_cfg),
        (r'/datasets/(?P<dataset_id>\w+)/logs', DatasetMultiLogsHandler, handler_cfg),
        (r'/datasets/(?P<dataset_id>\w+)/logs/(?P<log_id>\w+)', DatasetLogsHandler, handler_cfg),
        (r'/datasets/(?P<dataset_id>\w+)/tasks/(?P<task_id>\w+)/logs', DatasetTaskLogsHandler, handler_cfg),
    ]


class BaseHandler(RESTHandler):
    def initialize(self, database=None, s3=None, **kwargs):
        super(BaseHandler, self).initialize(**kwargs)
        self.db = database
        self.s3 = s3

class MultiLogsHandler(BaseHandler):
    """
    Handle logs requests.
    """
    @authorization(roles=['admin','client'])
    async def get(self):
        """
        Get multiple log entries based on search arguments.

        Body args (json):
            from (str): timestamp to start searching
            to (str): timestamp to end searching
            name (str): name (type) of log
            dataset_id (str): dataset_id
            task_id (str): task_id
            keys: | separated list of keys to return for each task

        Returns:
            dict: {log_id: {keys}}
        """
        query = {}
        projection = {'_id': False, 'log_id': True}
        try:
            limit = int(self.get_argument('limit', 0))
        except Exception:
            raise tornado.web.HTTPError(500, reason='non-integer limit')

        date_to = self.get_argument('to', None)
        date_from = self.get_argument('from', None)
        if date_to and date_from:
            query['timestamp'] = {'$gte': date_from, '$lte': date_to}
        elif date_from:
            query['timestamp'] = {'$gte': date_from}
        elif date_to:
            query['timestamp'] = {'$lte': date_to}

        for k in ('name', 'dataset_id', 'task_id'):
            val = self.get_argument(k, None)
            if val:
                query[k] = val

        keys = self.get_argument('keys', 'name|data|timestamp|task_id|dataset_id')
        projection.update({x:True for x in keys.split('|') if x})

        logging.debug('query: %r', query)
        logging.debug('projection: %r', projection)

        ret = {}
        f = self.db.logs.find(query, projection=projection, limit=limit)
        async for row in f:
            if 'data' in projection and 'data' not in row:
                if self.s3:
                    row['data'] = await self.s3.get(row['log_id'])
                else:
                    raise tornado.web.HTTPError(500, reason='no data field and s3 disabled')
            ret[row['log_id']] = row
        self.write(ret)
        self.finish()

    @authorization(roles=['admin','client','pilot'])
    async def post(self):
        """
        Create a log entry.

        Body should contain the following fields:
            data: str

        Optional fields:
            dataset_id: str
            task_id: str
            name: str

        Returns:
            dict: {'result': <log_id>}
        """
        data = json.loads(self.request.body)
        if 'data' not in data:
            raise tornado.web.HTTPError(400, reason='data field not in body')
        if 'name' not in data:
            data['name'] = 'log'
        log_id = uuid.uuid1().hex
        data['log_id'] = log_id
        data['timestamp'] = nowstr()
        if self.s3 and len(data['data']) > 1000000:
            await self.s3.put(log_id, data['data'])
            del data['data']
        ret = await self.db.logs.insert_one(data)
        self.set_status(201)
        self.write({'result': log_id})
        self.finish()

class LogsHandler(BaseHandler):
    """
    Handle logs requests.
    """
    @authorization(roles=['admin'])
    async def get(self, log_id):
        """
        Get a log entry.

        Args:
            log_id (str): the log id of the entry

        Returns:
            dict: all body fields
        """
        ret = await self.db.logs.find_one({'log_id':log_id},
                projection={'_id':False})
        if not ret:
            self.send_error(404, reason="Log not found")
        else:
            if 'data' not in ret:
                if self.s3:
                    ret['data'] = await self.s3.get(ret['log_id'])
                else:
                    raise tornado.web.HTTPError(500, reason='no data field and s3 disabled')
            self.write(ret)
            self.finish()

    @authorization(roles=['admin','client'])
    async def delete(self, log_id):
        """
        Delete a log entry.

        Args:
            log_id (str): the log id of the entry

        Returns:
            dict: empty dict on success
        """
        ret = await self.db.logs.find_one_and_delete({'log_id':log_id})
        if ret:
            if 'data' not in ret:
                if self.s3:
                    e = await self.s3.exists(log_id)
                    if e:
                        await self.s3.delete(log_id)
                else:
                    logging.warn('no data field and s3 disabled')
        self.write({})
        self.finish()

class DatasetMultiLogsHandler(BaseHandler):
    """
    Handle logs requests.
    """
    @authorization(roles=['admin'], attrs=['dataset_id:write'])
    async def post(self, dataset_id):
        """
        Create a log entry.

        Body should contain the following fields:
            data: str

        Optional fields:
            task_id: str
            name: str

        Args:
            dataset_id (str): the dataset id

        Returns:
            dict: {'result': <log_id>}
        """
        data = json.loads(self.request.body)
        if 'data' not in data:
            raise tornado.web.HTTPError(400, reason='data field not in body')
        log_id = uuid.uuid1().hex
        data['log_id'] = log_id
        data['dataset_id'] = dataset_id
        data['timestamp'] = nowstr()
        if self.s3 and len(data['data']) > 1000000:
            await self.s3.put(log_id, data['data'])
            del data['data']
        ret = await self.db.logs.insert_one(data)
        self.set_status(201)
        self.write({'result': log_id})
        self.finish()

class DatasetLogsHandler(BaseHandler):
    """
    Handle logs requests.
    """
    @authorization(roles=['admin'], attrs=['dataset_id:read'])
    async def get(self, dataset_id, log_id):
        """
        Get a log.

        Args:
            dataset_id (str): the dataset id
            log_id (str): the log id of the entry

        Returns:
            dict: all body fields
        """
        ret = await self.db.logs.find_one({'dataset_id':dataset_id,'log_id':log_id},
                projection={'_id':False})
        if not ret:
            self.send_error(404, reason="Log not found")
        else:
            if 'data' not in ret:
                if self.s3:
                    ret['data'] = await self.s3.get(ret['log_id'])
                else:
                    raise tornado.web.HTTPError(500, reason='no data field and s3 disabled')
            self.write(ret)
            self.finish()

class DatasetTaskLogsHandler(BaseHandler):
    """
    Handle log requests for a task
    """
    @authorization(roles=['admin'], attrs=['dataset_id:read'])
    async def get(self, dataset_id, task_id):
        """
        Get logs for a dataset and task.

        Note: "num" and "group" are generally not used together.

        Params (optional):
            num (int): number of logs, or groups of logs, to return
            group {true, false}: group by log name
            order {asc, desc}: order by time
            keys: | separated list of keys to return for each log

        Args:
            dataset_id (str): the dataset id
            task_id (str): the task id

        Returns:
            dict: {'logs': [log entry dict, log entry dict]}
        """
        filters = {'dataset_id': dataset_id, 'task_id': task_id}

        num = self.get_argument('num', None)
        if num:
            try:
                num = int(num)
            except Exception:
                raise tornado.web.HTTPError(400, reason='bad num param. must be int')

        group = self.get_argument('group', 'false').lower() == 'true'
        order = self.get_argument('order', 'desc').lower()
        if order not in ('asc', 'desc'):
            raise tornado.web.HTTPError(400, reason='bad order param. should be "asc" or "desc".')
        
        projection = {'_id': False}
        keys = self.get_argument('keys', None)
        if keys:
            projection.update({x:True for x in keys.split('|') if x})
        
        steps = [
            {'$match': filters},
            {'$sort': {'timestamp': -1 if order == 'desc' else 1}},
        ]
        if group:
            if not keys:
                keys = 'log_id|name|task_id|dataset_id|data|timestamp'
            grouping = {x:{'$first':'$'+x} for x in keys.split('|') if x}
            grouping['_id'] = '$name'
            if 'timestamp' not in grouping:
                grouping['timestamp'] = {'$first': '$timestamp'}
            steps.extend([
                {'$group': grouping},
                {'$sort': {'timestamp': -1 if order == 'desc' else 1}},
            ])
        steps.append({'$project': projection})
        if num:
            steps.append({'$limit': num})
        logger.debug('steps: %r', steps)

        cur = self.db.logs.aggregate(steps, allowDiskUse=True)
        ret = []
        async for entry in cur:
            ret.append(entry)
        if not ret:
            self.send_error(404, reason="Log not found")
        else:
            for log in ret:
                if 'data' not in log and ((not keys) or 'data' in keys.split('|')):
                    try:
                        if self.s3:
                            log['data'] = await self.s3.get(log['log_id'])
                        else:
                            raise Exception('no data field and s3 disabled')
                    except Exception as e:
                        self.send_error(500, reason=str(e))
                        return
            self.write({'logs':ret})
            self.finish()

    @authorization(roles=['admin'], attrs=['dataset_id:write'])
    async def delete(self, dataset_id, task_id):
        """
        Delete all logs for a dataset and task.

        Args:
            dataset_id (str): the dataset id
            task_id (str): the task id

        Returns:
            dict: empty dict on success
        """
        async for row in self.db.logs.find({'dataset_id': dataset_id, 'task_id': task_id}):
            log_id = row['log_id']
            await self.db.logs.delete_one({'log_id':log_id})
            if 'data' not in row:
                if self.s3:
                    e = await self.s3.exists(log_id)
                    if e:
                        await self.s3.delete(log_id)
                else:
                    logging.warn('no data field and s3 disabled')
        self.write({})
        self.finish()



"""
The REST API module runs any endpoints in the REST API,
as specified in the configuration.
"""
import logging
import importlib
import os

from rest_tools.server import RestServer

import iceprod.server
from iceprod.server import module
from iceprod.server import get_pkgdata_filename

logger = logging.getLogger('modules_rest_api')

class rest_api(module.module):
    """
    Run the REST API module, which handles REST API endpoints.

    Endpoints live in individual modules in `iceprod.server.rest_api.*`.
    """
    def __init__(self,*args,**kwargs):
        super(rest_api,self).__init__(*args,**kwargs)

        # set up the REST API
        routes, args = setup_rest(self.cfg, module=self)
        self.server = RestServer(**args)
        for r in routes:
            self.server.add_route(*r)

        kwargs = {}
        if 'rest_api' in self.cfg:
            if 'address' in self.cfg['rest_api']:
                kwargs['address'] = self.cfg['rest_api']['address']
            if 'port' in self.cfg['rest_api']:
                kwargs['port'] = self.cfg['rest_api']['port']
        self.server.startup(**kwargs)


def setup_rest(config, module=None):
    """
    Setup a REST Tornado server according to the config.

    The config should have defined the REST apis to use::

        {"rest": {"name1":{...}, "name2":{...} } }

    Args:
        config (:py:class:`iceprod.server.config.Config`): An IceProd config

    Returns:
        tuple: (routes, application args)
    """
    rest = config.get('rest',{})
    routes = []
    for key in rest:
        logger.warning('setting up tornado for REST.%s', key)
        mod = importlib.import_module('iceprod.server.rest.'+key)
        routes.extend(mod.setup(config, module=module))
    logger.info('REST routes being served:')
    for r in routes:
        logger.info('  %r', r)

    # get package data
    static_path = get_pkgdata_filename('iceprod.server','data/www')
    if static_path is None or not os.path.exists(static_path):
        logger.info('static path: %r',static_path)
        raise Exception('bad static path')
    template_path = get_pkgdata_filename('iceprod.server','data/www_templates')
    if template_path is None or not os.path.exists(template_path):
        logger.info('template path: %r',template_path)
        raise Exception('bad template path')

    kwargs = {
        'static_path': static_path,
        'template_path': template_path,
        'max_body_size': 10**9, # 1GB
    }
    return (routes, kwargs)